"""
Module for interacting with the Ed API.
"""

import json
import os
from typing import NoReturn, Optional
from datetime import datetime
import requests
from requests.compat import urljoin

from edapi.types.api_types.endpoints.analytics import API_Analytics_Users_Response

from .types import EdAuthError, EdError, EditThreadParams, PostThreadParams
from .types.api_types.endpoints.activity import (
    API_ListUserActivity_Response,
    API_ListUserActivity_Response_Item,
)
from .types.api_types.endpoints.files import API_PostFile_Response
from .types.api_types.endpoints.threads import (
    API_GetThread_Response,
    API_ListThreads_Response,
    API_PostThread_Response,
    API_PutThread_Request,
    API_PutThread_Response,
    API_PutThread_Response_Thread,
)
from .types.api_types.endpoints.user import API_User_Response
from .types.api_types.thread import API_Thread_WithComments, API_Thread_WithUser
from .types.api_types.user import API_User_WithEmail
from .models.thread import Thread, ThreadWithComments
from .models.user import User


def ANSI_BLUE(text): return f"\u001b[34m{text}\u001b[0m"
def ANSI_GREEN(text): return f"\u001b[32m{text}\u001b[0m"
def ANSI_RED(text): return f"\u001b[31m{text}\u001b[0m"


STATIC_FILE_BASE_URL = "https://static.us.edusercontent.com/files/"

API_TOKEN_ENV_VAR = "ED_API_TOKEN"

AUTH_MESSAGE = f"""
Go to
    {ANSI_BLUE("https://edstem.org/us/settings/api-tokens")}
and create a new API token.

Create a {ANSI_BLUE(".env")} file (or set up environment variables) with the {API_TOKEN_ENV_VAR} environment variable set to the API token you just created.
"""


def _ensure_login(func):
    """
    Decorator to ensure the user is logged in before calling the function.
    """

    def login_wrapper(self, *args, **kwargs):
        if self.api_token is None:
            if self.interactive:
                self.login()
            else:
                # no interaction; throw error instead.
                raise EdAuthError({"message": "No API token was provided"})
        return func(self, *args, **kwargs)

    return login_wrapper


def _throw_error(message: str, error_content: bytes) -> NoReturn:
    """
    Throw an error with the given message and the error content.
    """
    error_json = json.loads(error_content)
    if error_json.get("code", None) == "bad_token":
        # auth error
        raise EdAuthError(
            {"message": f"Ed authentication error; {message}", "response": error_json}
        )

    # other error
    raise EdError({"message": message, "response": error_json})


class EdAPI:
    """
    Class for Ed API integration.

    This class is responsible for authenticating the user, and for making API calls to the Ed API.
    """

    def __init__(self, api_token: str, base_url="https://edstem.org/api/", interactive=True):
        """
        Initialize the Ed API.

        If `api_token` is specified here, no attempt will be made
        to load a token from a `.env` file.

        By default, `interactive=True`, and the user will be prompted
        if no valid API token is found.
        If `interactive` is set to `False`, no such prompt will occur,
        and methods will raise an error.

        The `login()` method can still be called for interactive behavior.

        """
        if api_token is None:
            _throw_error("No API token was provided.", b"")
        self.api_token = api_token
        self.interactive = interactive
        self.api_base_url = base_url
        self.session = requests.Session()
        self.session.headers.update(self._auth_header)
        try:
            self.user_info = self.get_user_info()
        except EdAuthError:
            _throw_error("Invalid API token.", b"")

    def login(self):
        """
        Log in to the Ed API with the API token.

        Continuously prompts for the API token if it is not found in the .env file.
        """
        user_info = None
        while self.api_token is None:
            print(AUTH_MESSAGE)
            input(ANSI_RED("Press ENTER when you have done so."))
            print()  # new line to divide next output
            user_info = self._load_api_token()

            if self.api_token is None:
                print(
                    ANSI_RED(
                        "Invalid API Token; make sure you have the correct token.")
                )

        if user_info is None:
            # only way this is reached is if the token was initially loaded successfully.
            user_info: User = self.get_user_info()

        # print login welcome message
        user_name = user_info.name
        user_email = user_info.email
        print(
            "Authentication successful;",
            f"logged in as {ANSI_GREEN(user_name)} ({ANSI_BLUE(user_email)})",
        )

    @property
    def _auth_header(self):
        """
        Auth header with the API token for all requests.
        """
        return {"Authorization": f"Bearer {self.api_token}"}

    def _remove_auth_header(self):
        """
        Remove the auth header from the session.
        """
        del self.session.headers["Authorization"]

    @_ensure_login
    def get_user_info(self) -> User:
        """
        Retrieve the user info from Ed.
        """
        user_info_url = urljoin(self.api_base_url, "user")
        response = self.session.get(user_info_url)
        if response.ok:
            response_json: API_User_Response = response.json()
            return User.from_api_response(response_json)

        _throw_error("Failed to get user info.", response.content)

    @_ensure_login
    def list_user_activity(
        self,
        /,
        user_id: int,
        course_id: int,
        *,
        limit: int = 30,
        offset: int = 0,
        filter: str = "all",  # pylint: disable=redefined-builtin
    ) -> list[API_ListUserActivity_Response_Item]:
        """
        Retrieve a list of comments and threads made by the user.

        Limit can range from 1 to 50 (anything higher will get clipped to 50).
        Offset can be used to list out all activity for a user iteratively,
        through pagination.

        GET /api/users/<user_id>/profile/activity?courseID=<course_id>
        """
        list_url = urljoin(self.api_base_url,
                           f"users/{user_id}/profile/activity")
        response = self.session.get(
            list_url,
            params={
                "courseID": course_id,
                "limit": limit,
                "offset": offset,
                "filter": filter,
            },
        )
        if response.ok:
            response_json: API_ListUserActivity_Response = response.json()
            return response_json.get("items", [])  # default to empty list

        _throw_error(
            f"Failed to list user activity for user {user_id} in course {course_id}.",
            response.content,
        )

    @_ensure_login
    def list_threads(
        self, /, course_id: int, *, limit: int = 30, offset: int = 0, sort: str = "new"
    ) -> list[Thread]:
        """
        Retrieve list of threads, with the given limit, offset, and sort.

        Limit can range from 1 to 100 (anything higher will get clipped to 100).
        Offset can be used to list out all of the threads in a course iteratively,
        through pagination.

        GET /api/courses/<course_id>/threads
        """
        list_url = urljoin(self.api_base_url, f"courses/{course_id}/threads")
        response = self.session.get(
            list_url, params={"limit": limit, "offset": offset, "sort": sort}
        )
        if response.ok:
            response_json: API_ListThreads_Response = response.json()
            api_threads = response_json["threads"]
            return [Thread.model_validate(thread) for thread in api_threads]

        _throw_error(
            f"Failed to list threads for course {course_id}.", response.content
        )

    def list_last_updated_threads(self, course_id: int, date: datetime, limit=100, offset=0, sort="new") -> list[Thread]:
        """
        Retrieve list of threads, with the given limit, offset, and sort.

        Limit can range from 1 to 100 (anything higher will get clipped to 100).
        Offset can be used to list out all of the threads in a course iteratively,
        through pagination.

        GET /api/courses/<course_id>/threads
        """
        threads = self.list_threads(
            course_id, limit=limit, offset=offset, sort=sort)
        print(threads[0].updated_at)
        return [thread for thread in threads if thread.updated_at > date]

    def list_users(self, /, course_id: int) -> list[API_User_WithEmail]:
        """
        Retrieve list of users.

        GET /api/courses/<course_id>/analytics/users
        """
        list_url = urljoin(
            self.api_base_url, f"courses/{course_id}/analytics/users")
        response = self.session.get(list_url)
        if response.ok:
            response_json: API_Analytics_Users_Response = response.json()
            return response_json["users"]

        _throw_error(
            f"Failed to list users for course {course_id}", response.content)

    @_ensure_login
    def get_thread(self, thread_id: int) -> API_Thread_WithComments:
        """
        Retrieve the details for a thread, given its id.

        GET /api/threads/<thread_id>
        """
        thread_url = urljoin(self.api_base_url, f"threads/{thread_id}")
        response = self.session.get(thread_url)
        if response.ok:
            response_json: API_GetThread_Response = response.json()
            return response_json["thread"]

        _throw_error(f"Failed to get thread {thread_id}.", response.content)

    @_ensure_login
    def get_course_thread(
        self, course_id: int, thread_number: int
    ) -> ThreadWithComments:
        """
        Retrieve the details for a thread in a given course, using the thread number.

        GET /api/courses/<course_id>/threads/<thread_id>
        """
        thread_url = urljoin(
            self.api_base_url, f"courses/{course_id}/threads/{thread_number}"
        )
        response = self.session.get(thread_url)
        if response.ok:
            response_json: API_GetThread_Response = response.json()
            thread_api_res = response_json["thread"]
            return ThreadWithComments.model_validate(thread_api_res)

        _throw_error(
            f"Failed to get thread {thread_number}.", response.content)

    def get_course_threads_by_id(self, thread_id: int) -> ThreadWithComments:
        """
        Retrieve the details for a thread in a given course, using the thread number.

        GET /api/courses/<course_id>/threads/<thread_id>
        """
        thread_url = urljoin(self.api_base_url, f"threads/{thread_id}")
        response = self.session.get(thread_url)
        if response.ok:
            response_json: API_GetThread_Response = response.json()
            thread_api_res = response_json["thread"]
            return ThreadWithComments.model_validate(thread_api_res)

        _throw_error(
            f"Failed to get thread {thread_id}.", response.content)

    @_ensure_login
    def post_thread(
        self, course_id: int, params: PostThreadParams
    ) -> API_Thread_WithUser:
        """
        Creates a new thread in the given course.

        POST /api/courses/<course_id>/threads

        Returns newly created thread object.
        """
        thread_url = urljoin(self.api_base_url, f"courses/{course_id}/threads")
        response = self.session.post(thread_url, json={"thread": params})
        if response.ok:
            response_json: API_PostThread_Response = response.json()
            return response_json["thread"]

        _throw_error(
            f"Failed to post thread in course {course_id}.", response.content)

    @_ensure_login
    def edit_thread(
        self, thread_id: int, params: EditThreadParams, unlock_thread=True
    ) -> API_PutThread_Response_Thread:
        """
        Edit the details for a given thread.

        If `unlock_thread` is True (default), then the thread is unlocked
        before editing. Otherwise, if `unlock_thread` is False, no attempt
        to unlock the thread will be made, and the edit request will fail.

        PUT /api/threads/<thread_id>

        Returns newly created thread object.
        """

        thread = self.get_thread(thread_id)

        relock = False
        if unlock_thread and thread["is_locked"]:
            # locked thread, so unlock it and re-lock at the very end
            relock = True
            self.unlock_thread(thread_id)
            # fetch thread again in case any side-effects happen
            thread = self.get_thread(thread_id)

        # set items
        for key, val in params.items():
            # ensure we're only modifying values that have appeared in the existing thread object
            if key in thread and val is not None:
                thread[key] = val

        thread_url = urljoin(self.api_base_url, f"threads/{thread_id}")
        request_json: API_PutThread_Request = {"thread": thread}
        response = self.session.put(thread_url, json=request_json)
        if response.ok:
            response_json: API_PutThread_Response = response.json()

            if relock:
                # relock thread if necessary
                self.lock_thread(thread_id)

            return response_json["thread"]

        _throw_error(f"Failed to edit thread {thread_id}.", response.content)

    @_ensure_login
    def upload_file(self, filename: str, file: bytes, content_type: str) -> str:
        """
        Upload a file to Ed.

        `filename` is used as the filename for the file when uploading.
        `file` is the raw bytestream from `open(..., "rb")` after calling `.read()`.
        `content_type` is the MIME type of the file.

        POST /api/files

        Returns the static URL for the uploaded file.
        """
        upload_url = urljoin(self.api_base_url, "files")
        # send file through formdata
        formdata = {"attachment": (filename, file, content_type)}
        response = self.session.post(upload_url, files=formdata)
        if response.ok:
            response_json: API_PostFile_Response = response.json()
            file_id = response_json["file"]["id"]
            return urljoin(STATIC_FILE_BASE_URL, file_id)

        _throw_error(f"Failed to upload file {filename}.", response.content)

    @_ensure_login
    def lock_thread(self, thread_id: int) -> None:
        """
        Lock a given thread.

        POST /api/threads/<thread_id>/lock
        """
        lock_url = urljoin(self.api_base_url, f"threads/{thread_id}/lock")
        response = self.session.post(lock_url)
        if not response.ok:
            _throw_error(
                f"Failed to lock thread {thread_id}.", response.content)

    @_ensure_login
    def unlock_thread(self, thread_id: int) -> None:
        """
        Unlock a given thread.

        POST /api/threads/<thread_id>/unlock
        """
        unlock_url = urljoin(self.api_base_url, f"threads/{thread_id}/unlock")
        response = self.session.post(unlock_url)
        if not response.ok:
            _throw_error(
                f"Failed to unlock thread {thread_id}.", response.content)
